/*
 * Copyright (c) 2014-2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Thu Nov 20 2014
 */
#include "acheader.h"

#include <stdio.h>
#include <stdarg.h>

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/atomic.h>
#include <util/twi.h>

#include "tk3-mikrokopter.h"


/* --- local data ---------------------------------------------------------- */

struct tk3_iobuffer {
  volatile uint8_t recvbuf[64];
  volatile uint8_t recv_r, recv_w;
  volatile tk3_time recv_ts;

  volatile uint8_t sendbuf[64];
  volatile uint8_t send_r, send_w;
};

static struct tk3_iobuffer uart0;
static struct tk3_iobuffer twi;

static int	tk3_io_recv(struct tk3_iobuffer *io, struct tk3_iorecv *iorecv);
static void	tk3_encodechar(char c, struct tk3_iobuffer *io);
static void	tk3_putchar(char c, struct tk3_iobuffer *io);


/* --- tk3_uart_init ------------------------------------------------------- */

int
tk3_uart_init(int unit, uint32_t baud)
{
  uint16_t b;

  /* set baudrate */
  b = F_CPU / baud / 8 - 1;
  UBRR0H = (b >> 8) & 0xff;
  UBRR0L = b & 0xff;

  /* double clock speed */
  UCSR0A = (1 << U2X0);
  /* enable UART0 for transmission and reception */
  UCSR0B = (1 << RXCIE0) | (1 << RXEN0) | (1 << TXEN0);
  /* 8/N/1 */
  UCSR0C = (1 << UCSZ01) | (1 << UCSZ00);

  /* initialize buffers */
  uart0.recv_r = 0;
  uart0.recv_w = 0;
  uart0.send_r = 0;
  uart0.send_w = 0;

  return 0;
}


/* --- tk3_twi_init -------------------------------------------------------- */

int
tk3_twi_init(uint8_t addr)
{
  TWSR &= ~(( 1 << TWPS1) | (1 << TWPS0));
  TWBR = ((F_CPU / 400000) - 16) / (2 * 1);

  TWAR = (addr << 1) | (1 << TWGCE);
  TWCR = (1 << TWEA) | (1 << TWEN) | (1 << TWIE);

  /* initialize buffers */
  twi.recv_r = 0;
  twi.recv_w = 0;
  twi.send_r = 0;
  twi.send_w = 0;

  return 0;
}

/* --- tk3_log ------------------------------------------------------------- */

int
tk3_log(enum tk3_channel channel, const char *fmt, ...)
{
  struct tk3_iobuffer *io;
  va_list ap;
  char c;

  switch(channel) {
    case TK3_UART0: io = &uart0; break;
    case TK3_TWI: io = &twi; break;
    default: return 1;
  }

  va_start(ap, fmt);
  tk3_putchar('^', io);
  while((c = pgm_read_byte(fmt++))) {
    switch(c) {
      case '%': {
        switch(pgm_read_byte(fmt++)) {
          case '1':
            tk3_encodechar(va_arg(ap, int), io);
            break;

          case '2': {
            uint16_t x = va_arg(ap, uint16_t);
            tk3_encodechar((x >> 8) & 0xff, io);
            tk3_encodechar(x & 0xff, io);
            break;
          }

          case '4': {
            uint32_t x = va_arg(ap, uint32_t);
            tk3_encodechar((x >> 24) & 0xff, io);
            tk3_encodechar((x >> 16) & 0xff, io);
            tk3_encodechar((x >> 8) & 0xff, io);
            tk3_encodechar(x & 0xff, io);
            break;
          }
        }
        break;
      }

      default:
        tk3_encodechar(c, io);
    }
  }
  tk3_putchar('$', io);
  va_end(ap);
  return 0;
}


/* --- tk3_io_recv --------------------------------------------------------- */

struct tk3_iorecv *
tk3_recv(void)
{
  static struct tk3_iorecv io_uart0 = { .state.channel = TK3_UART0 };
  static struct tk3_iorecv io_twi = { .state.channel = TK3_TWI };

  if (tk3_io_recv(&uart0, &io_uart0)) return &io_uart0;
  if (tk3_io_recv(&twi, &io_twi)) return &io_twi;
  return NULL;
}

static int
tk3_io_recv(struct tk3_iobuffer *io,  struct tk3_iorecv *iorecv)
{
  uint8_t r;
  char c;

  r = io->recv_r;
  while(r != io->recv_w) {
    c = io->recvbuf[r];
    r = (r + 1) % sizeof(io->recvbuf);
    io->recv_r = r;

#ifdef LEGACY_MKFL
    /* XXX legacy FL is just sending 8 bits motor PWM */
    /* XXX will be dropped shortly */
    if (iorecv->state.channel == TK3_TWI) {
      iorecv->data[0] = c;
      iorecv->len = 1;
      return 1;
    }
#endif

    switch (c) {
      case '^':
        iorecv->state.start = 1;
        iorecv->state.escape = 0;
        iorecv->ts = io->recv_ts;
        iorecv->len = 0;
        break;

      case '$':
        if (!iorecv->state.start) break;
        iorecv->state.start = 0;
        return 1;

      case '!':
        iorecv->state.start = 0;
        break;

      case '\\':
        iorecv->state.escape = 1;
        break;

      default:
        if (!iorecv->state.start) break;
        if (iorecv->len >= sizeof(iorecv->data)) {
          iorecv->state.start = 0;
          break;
        }
        if (iorecv->state.escape) {
          iorecv->state.escape = 0;
          c = ~c;
        }

        iorecv->data[iorecv->len++] = c;
        break;
    }
  }

  return 0;
}


/* --- tk3_io_send --------------------------------------------------------- */

void
tk3_send()
{
  uint8_t r;
  char c;

  /* twi */
  if (twi.send_r != twi.send_w) {
    if (TW_STATUS == TW_NO_INFO) {
      TWCR =
        (1 << TWINT) | (1 << TWEA) | (1 << TWSTA) | (1 << TWEN) | (1 << TWIE);
    }
  }

  /* uart */
  r = uart0.send_r;
  while(r != uart0.send_w) {
    c = uart0.sendbuf[r];
    r = (r + 1) % sizeof(uart0.sendbuf);
    uart0.send_r = r;

    while(!(UCSR0A & (1 << UDRE0))) /* empty body */;
    UDR0 = c;
  }
}


/* --- tk3_encodechar ------------------------------------------------------ */

static void
tk3_encodechar(char c, struct tk3_iobuffer *io)
{
  switch (c) {
    case '^': case '$':  case '!': case '\\':
      tk3_putchar('\\', io);
      c = ~c;
      break;
  }
  tk3_putchar(c, io);
}


/* --- tk3_putchar --------------------------------------------------------- */

static void
tk3_putchar(char c, struct tk3_iobuffer *io)
{
  uint8_t w, r;

  w = io->send_w;
  io->sendbuf[w] = c;

  w = (w + 1) % sizeof(io->sendbuf);
  r = io->send_r;
  if (w == r) {
    r = (r + 1) % sizeof(io->sendbuf);
    io->sendbuf[r] = '!';
    io->send_r = r;
  }
  io->send_w = w;
}


/* --- USART0_RX_vect ------------------------------------------------------ */

#ifndef USART0_RX_vect
# ifdef USART_RX_vect
#   define USART0_RX_vect USART_RX_vect
# endif
#endif

ISR(USART0_RX_vect)
{
  static uint8_t tsmagic;

  uint8_t status = UCSR0A;
  uint8_t c = UDR0;
  uint8_t w, r;

  if (status & ((1 << FE0) | (1 << UPE0) | (1 << DOR0)))
    c = '!';

  /* get local timestamp of the last received timestamp message */
  if (tsmagic) {
    if (c == 't')
      uart0.recv_ts = tk3_clock_gettime();
    tsmagic = 0;
  } else {
    if (c == '^') tsmagic = 1;

    /* wake up sleeping activities */
    if (c == '$') tk3_clock_interrupt_delay();
  }

  w = uart0.recv_w;
  r = uart0.recv_r;
  uart0.recvbuf[w] = c;
  w = (w + 1) % sizeof(uart0.recvbuf);
  if (w == r) {
    /* buffer overrun: discard old data and make sure next read char is '!' */
    r = (r + 1) % sizeof(uart0.recvbuf);
    uart0.recvbuf[r] = '!';
    uart0.recv_r = r;
#ifdef TK3_IOBUF_OVERFLOW
    TK3_IOBUF_OVERFLOW;
#endif
  } else {
#ifdef TK3_IOBUF_NOVERFLOW
    TK3_IOBUF_NOVERFLOW;
#endif
  }

  uart0.recv_w = w;
}


/* --- TWI_vect ------------------------------------------------------------ */

ISR(TWI_vect)
{
  static uint8_t tsmagic;
  static int gccmd;

  uint8_t c, w, r;

#define TWCR_CLRINT ((1 << TWINT) | (1 << TWEA) | (1 << TWEN) | (1 << TWIE))

  switch (TW_STATUS) {
    /* slave receiver mode */
    case TW_SR_ARB_LOST_SLA_ACK:
    case TW_SR_ARB_LOST_GCALL_ACK:
      /* arbitration lost in SLA+RW, SLA+W received, ACK returned */
      /* or arbitration lost in SLA+RW, general call received, ACK returned */
      /* FALLTHROUGH */
    case TW_SR_SLA_ACK:
    case TW_SR_GCALL_ACK:
      /* SLA+W received, ACK returned */
      /* or general call received, ACK returned */
      gccmd = 0;
      TWCR = TWCR_CLRINT;
      break;

    case TW_SR_GCALL_DATA_ACK:
    case TW_SR_GCALL_DATA_NACK:
      /* general call data received, ACK returned */
      /* or general call data received, NACK returned */
      if (!gccmd) {
        gccmd = TWDR;
        TWCR = TWCR_CLRINT;
        break;
      }
      /* FALLTHROUGH */
    case TW_SR_DATA_ACK:
    case TW_SR_DATA_NACK:
      /* data received, ACK returned */
      /* or data received, NACK returned */
      c = TWDR;

      /* get local timestamp of the last received timestamp message */
      if (tsmagic) {
        if (c == 't')
          twi.recv_ts = tk3_clock_gettime();
        tsmagic = 0;
      } else {
        if (c == '^') tsmagic = 1;

        /* wake up sleeping activities */
        if (c == '$') tk3_clock_interrupt_delay();
      }

#ifdef LEGACY_MKFL
      /* XXX legacy FL is just sending 8 bits motor PWM without end-of-message
       * marker */
      /* XXX will be dropped shortly */
      tk3_clock_interrupt_delay();
#endif

      w = twi.recv_w;
      r = twi.recv_r;
      twi.recvbuf[w] = c;
      w = (w + 1) % sizeof(twi.recvbuf);
      if (w == r) {
        /* buffer overrun: discard old data and make sure next char is '!' */
        r = (r + 1) % sizeof(uart0.recvbuf);
        twi.recvbuf[r] = '!';
        twi.recv_r = r;
#ifdef TK3_IOBUF_OVERFLOW
        TK3_IOBUF_OVERFLOW;
#endif
      } else {
#ifdef TK3_IOBUF_NOVERFLOW
        TK3_IOBUF_NOVERFLOW;
#endif
      }
      twi.recv_w = w;

      TWCR = TWCR_CLRINT;
      break;

    case TW_SR_STOP:
      /* stop or repeated start condition received while selected */
      if (twi.send_w != twi.send_r)
        TWCR = TWCR_CLRINT | (1 << TWSTA);
      else {
        TWCR = TWCR_CLRINT;
      }
      break;


      /* master transmitter mode */
    case TW_START:
    case TW_REP_START:
      /* start condition transmitted */
      /* or repeated start condition transmitted */
      TWDR = 0 /* general call */;
      TWCR = TWCR_CLRINT;
      break;

    case TW_MT_SLA_ACK:
    case TW_MT_SLA_NACK:
      /* SLA+W transmitted, ACK received */
      /* or SLA+W transmitted, NACK received */

      TWDR = 1; /* general call command: own address<<1 | 1 */
      TWCR = TWCR_CLRINT;
      break;

    case TW_MT_DATA_ACK:
    case TW_MT_DATA_NACK:
      /* data transmitted, ACK received */
      /* data transmitted, NACK received */
      r = twi.send_r;
      if (r != twi.send_w) {
        TWDR = twi.sendbuf[r];
        r = (r + 1) % sizeof(twi.sendbuf);
        twi.send_r = r;

        TWCR = TWCR_CLRINT;
      } else {
        TWCR = TWCR_CLRINT | (1 << TWSTO);
      }
      break;

    case TW_MT_ARB_LOST:
      /* arbitration lost in SLA+W or data */
      TWCR = TWCR_CLRINT | (1 << TWSTA);
      break;

    case TW_NO_INFO:
      /* no state information available */
      break;

    case TW_BUS_ERROR:
      /* illegal start or stop condition */
      TWCR = TWCR_CLRINT | (1 << TWSTO);
      break;
  }
}
