/*
 * Copyright (c) 2015 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri Jan 30 2015
 */
#include "acheader.h"

#include <avr/pgmspace.h>
#include <util/atomic.h>

#include "tk3-mikrokopter.h"
#include "mkbl.h"


/* --- tk3_log_velocity ---------------------------------------------------- */

void
tk3_log_velocity(enum tk3_channel channel)
{
  uint8_t state;
  uint16_t period;

  state = 0;
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    if (tk3_motor.starting) state |= 1<<0;
    if (tk3_motor.spinning) state |= 1<<1;
    if (tk3_motor.servo) state |= 1<<2;
    if (tk3_motor.emerg) state |= 1<<7;
    period = tk3_motor.period;
  }

  if (period > tk3_motor.max_period)
    period = 0xffff;
  else
    period *= settings.motor_period_mul;

  tk3_log(channel, PSTR("S%1%2"), state, period);
}


/* --- tk3_log_current ----------------------------------------------------- */

void
tk3_log_current(enum tk3_channel channel)
{
  uint16_t current;

  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    current = tk3_sensors.current;
  }

  tk3_log(channel, PSTR("A%2"), current);
}


/* --- tk3_log_motor ------------------------------------------------------- */

void
tk3_log_motor(enum tk3_channel channel)
{
  tk3_time time;
  uint8_t state;
  uint16_t period, pwm, current;

  state = 0;
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    time = tk3_motor.stamp;
    if (tk3_motor.starting) state |= 1<<0;
    if (tk3_motor.spinning) state |= 1<<1;
    if (tk3_motor.servo) state |= 1<<2;
    if (tk3_motor.emerg) state |= 1<<7;
    period = tk3_motor.period;
    pwm = tk3_motor.pwm;
    current = tk3_sensors.peak_current;
    tk3_sensors.peak_current = tk3_sensors.current;
  }

  if (period > tk3_motor.max_period)
    period = 0xffff;
  else
    period *= settings.motor_period_mul;

  tk3_log(channel, PSTR("M%4%1%2%2%2"), time, state, pwm, period, current);
}


/* --- tk3_log_sensors ----------------------------------------------------- */

void
tk3_log_sensors(enum tk3_channel channel)
{
  tk3_time time;
  uint16_t battery, current, mcu, pcb;

  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    time = tk3_sensors.stamp;
    battery = tk3_sensors.battery;
    current = tk3_sensors.current;
    mcu = tk3_sensors.mcu_temp;
    pcb = tk3_sensors.pcb_temp;
  }

  tk3_log(channel, PSTR("D%4%2%2%2%2"), time, battery, current, mcu, pcb);
}


/* --- tk3_log_control ----------------------------------------------------- */

void
tk3_log_control(enum tk3_channel channel)
{
  tk3_time time;
  uint8_t state;
  uint16_t target;
  int16_t bias, gain, err;

  state = 0;
  ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
    time = tk3_motor.stamp;
    if (tk3_motor.starting) state |= 1<<0;
    if (tk3_motor.spinning) state |= 1<<1;
    if (tk3_motor.servo) state |= 1<<2;
    if (tk3_motor.emerg) state |= 1<<7;
    target = tk3_motor.target;
    bias = tk3_motor.smc_bias;
    gain = tk3_motor.smc_gain;
    err = tk3_motor.smc_err & 0xffff;
  }

  if (target > tk3_motor.max_period)
    target = 0xffff;
  else
    target *= settings.motor_period_mul;

  tk3_log(channel, PSTR("K%4%1%2%2%2%2"), time, state, target, bias, gain, err);
}
