/*
 * Copyright (c) 2014 LAAS/CNRS
 * Copyright (c) 2008-2014 Max Planck Institute for Biological Cybernetics
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 */
#include "acheader.h"

#include "main.h"
#include "nhfc.h"
#include "timer0.h"
#include "data_struct.h"
#include "mymath.h"
#include "twi.h"

#define ACC_FILT_GAIN 1
#define GYRO_OFFSET   509 // 509 // reading without rotation
#define GYRO_ZERO_VAL 515 // max reading (1024) - GYRO_OFFSET
#define GYRO_SCALING    4000000 // to have decimals of degrees/10000 (600 * 10000)
#define GYRO_MULT_NUM 5
#define GYRO_MULT_DEN	3
#define OMEGA_CUTOFF 100 //100 Hz
#define DRIFT_CUTOFF 1 // 1 Hz
#define ATTIT_EST_TIME 5 // time in 100um

#define KP_ATT_MULT 100
#define KD_ATT_MULT 10
#define KI_ATT_MULT 100
#define KP_YAW_MULT 10
#define KP_YAW_ACC_MULT 10

#define MIN_MOT_CMD 8
#define MAX_MOT_CMD 230

#define LP_FILTER_TRESH 2000


#define CHECK_MIN_MAX(data,min,max) {if(data < min) data = min; else if(data > max) data = max;}
#define SIGNUM(data) (data < 0 ? -1 : 1)

struct str_NearHovCmd nearHovCmds = { 0, 0, 0 , 0};

void nearHovEstUpdate(int reset){

	stateEstSteps++;

	const signed long int timeDiff_est = estPerMsDecs;

		if(timeDiff_est > ATTIT_EST_TIME){

			signed int diffMs = (timeDiff_est*103)/10;
			if(diffMs > mikKopData.maxFCPeriodMs){
				mikKopData.maxFCPeriodMs = diffMs;
			}

			signed long int estRollVel= (-(mikKopData.rawGyroX - GYRO_OFFSET) * GYRO_SCALING) / GYRO_ZERO_VAL; //don't use the drift value during the estimation
			signed long int estPitchVel= (-(mikKopData.rawGyroY - GYRO_OFFSET) * GYRO_SCALING) / GYRO_ZERO_VAL;
			signed long int estYawVel = ((mikKopData.rawGyroZ - GYRO_OFFSET) * GYRO_SCALING) / GYRO_ZERO_VAL;

			if (mikKopData.driftEstimActive){

				static signed long int driftRoll = 0, driftPitch = 0, driftYaw = 0;

				driftRoll = driftRoll + (2*DRIFT_CUTOFF*(-estRollVel-driftRoll)*timeDiff_est) / (3255);
				driftPitch = driftPitch + (2*DRIFT_CUTOFF*(-estPitchVel-driftPitch)*timeDiff_est) / (3255);
				driftYaw = driftYaw + (2*DRIFT_CUTOFF*(-estYawVel-driftYaw)*timeDiff_est) / (3255);

				// Drifts are rounded to the closest integer
				mikKopData.driftGyroX = (signed int) ((driftRoll + SIGNUM(driftRoll)*5)/10);
				mikKopData.driftGyroY = (signed int) ((driftPitch + SIGNUM(driftPitch)*5)/10);
				mikKopData.driftGyroZ = (signed int) ((driftYaw + SIGNUM(driftYaw)*5)/10);

			} else {

				// Correct omega with estimated drifts
				estRollVel += 10 * (long signed int)mikKopData.driftGyroX;
				estPitchVel += 10 * (long signed int)mikKopData.driftGyroY;
				estYawVel += 10 * (long signed int)mikKopData.driftGyroZ;

				// Apply scaling to have consistency with vicon
				estRollVel = GYRO_MULT_NUM*estRollVel/GYRO_MULT_DEN;
				estPitchVel = GYRO_MULT_NUM*estPitchVel/GYRO_MULT_DEN; //scaling to have consistency with vicon
				estYawVel = GYRO_MULT_NUM*estYawVel/GYRO_MULT_DEN;

				//printf("estYawVel=%10ld, estPitchVel=%10ld, estRollVel=%10ld ", estYawVel, estPitchVel, estRollVel);

				// Start attitude estimation
				static signed long int estRoll=0, estPitch=0; // static because are states
				static signed long int rollRateLowPass=0, pitchRateLowPass=0, yawRateLowPass=0;

				//------------- ROLL -------------//
				const signed long int rollDot_acc = c_asin(mikKopData.rawAccY - mikKopData.offsetRawAccY) - estRoll;
				const signed long int rollDot = ACC_FILT_GAIN*rollDot_acc + estRollVel;

				estRoll = estRoll + (rollDot * timeDiff_est) / 9700; //10000 parts of degrees

				//------------- PITCH -------------//
				const signed long int pitchDot_acc = -c_asin(mikKopData.rawAccX - mikKopData.offsetRawAccX) - estPitch;
				const signed long int pitchDot = ACC_FILT_GAIN*pitchDot_acc + estPitchVel;

				estPitch = estPitch + (pitchDot * timeDiff_est) / 9700; //10000 parts of degrees

			    // first-order low pass filtering of the angular velocity, to be used in the control loop, OMEGA_CUTOFF is in Hz
/*				rollRateLowPass += (2*3*OMEGA_CUTOFF*(estRollVel - rollRateLowPass))/485;
				pitchRateLowPass += (2*3*OMEGA_CUTOFF*(estPitchVel - pitchRateLowPass))/485;
				yawRateLowPass += (2*3*OMEGA_CUTOFF*(estYawVel - yawRateLowPass))/485;*/

				// TODO: note that 2*3*100/9765~= 2^-4 but maybe less efficient. One could do:
				// rollRateLowPass += ((estRollVel-rollRateLowPass)*timeDiff_est) >> 4

				// Reset and recovery logic
				if (reset) {
					rollRateLowPass = estRollVel;
					pitchRateLowPass = estPitchVel;
					yawRateLowPass = estYawVel;
				} else {
					signed long int tempDiff;

					tempDiff=(estRollVel - rollRateLowPass);
					if (tempDiff>LP_FILTER_TRESH || tempDiff< -LP_FILTER_TRESH){
						rollRateLowPass = estRollVel;
					}
					else {
						rollRateLowPass += ((((OMEGA_CUTOFF/10)*(estRollVel - rollRateLowPass))*timeDiff_est)/163);
					}

					tempDiff=(estPitchVel - pitchRateLowPass);
					if (tempDiff>LP_FILTER_TRESH || tempDiff< -LP_FILTER_TRESH){
						pitchRateLowPass = estPitchVel;
					}
					else {
						pitchRateLowPass += ((((OMEGA_CUTOFF/10)*(estPitchVel - pitchRateLowPass))*timeDiff_est)/163);
					}

					tempDiff=(estYawVel - yawRateLowPass);
					if (tempDiff>LP_FILTER_TRESH || tempDiff< -LP_FILTER_TRESH){
						yawRateLowPass = estYawVel;
					}
					else {
						yawRateLowPass += ((((OMEGA_CUTOFF/10)*(estYawVel - yawRateLowPass))*timeDiff_est)/163);
					}

					rollRateLowPass += ((((OMEGA_CUTOFF/10)*(estRollVel - rollRateLowPass))*timeDiff_est)/163); // HIC SUNT LEONES
					pitchRateLowPass += ((((OMEGA_CUTOFF/10)*(estPitchVel - pitchRateLowPass))*timeDiff_est)/163);
					yawRateLowPass += ((((OMEGA_CUTOFF/10)*(estYawVel - yawRateLowPass))*timeDiff_est)/163);
				}

				// Was before:
				//rollRateLowPass += 2*3*(((OMEGA_CUTOFF*(estRollVel - rollRateLowPass))*timeDiff_est)/9765);
				//pitchRateLowPass += 2*3*(((OMEGA_CUTOFF*(estPitchVel - pitchRateLowPass))*timeDiff_est)/9765);
				//yawRateLowPass += 2*3*(((OMEGA_CUTOFF*(estYawVel - yawRateLowPass))*timeDiff_est)/9765);

/*				rollRateLowPass += 2*3*((((estRollVel - rollRateLowPass))*timeDiff_est)/97);
				pitchRateLowPass += 2*3*((((estPitchVel - pitchRateLowPass))*timeDiff_est)/97);
				yawRateLowPass += 2*3*((((estYawVel - yawRateLowPass))*timeDiff_est)/97);*/

				// Copy data in mikKopData structure
				mikKopData.pitch = (signed int) (estPitch/1000); //dec of deg
				mikKopData.roll  = (signed int) (estRoll/1000); //dec of deg

				mikKopData.rollRate = (signed int) (estRollVel/1000); //dec of deg/sec
				mikKopData.pitchRate = (signed int) (estPitchVel/1000); //dec of deg/sec
				mikKopData.yawRate = (signed int) (estYawVel/1000); //dec of deg/sec

				mikKopData.rollRateLowPass  = (signed int) (rollRateLowPass/1000); //dec of deg/sec
				mikKopData.pitchRateLowPass = (signed int) (pitchRateLowPass/1000); //dec of deg/sec
				mikKopData.yawRateLowPass   = (signed int) (yawRateLowPass/1000); //dec of deg/sec

/*				printf("dt=%8ld, wX=%8d, wY=%8d, wZ=%8d, roll=%8d pitch=%8d\r", timeDiff_est,
						mikKopData.rollRateLowPass, mikKopData.pitchRateLowPass, mikKopData.yawRateLowPass,
						mikKopData.roll/10, mikKopData.pitch/10);*/
			}
		}

		estPerMsDecs = 0;
}




void nearHovCtrlStep(){

	//printf("nearHovCtrlStep\n\r");

	const int desPitch = (int) nearHovCmds.pitch;
	const int desRoll = (int) nearHovCmds.roll;
	const long desThrust = nearHovCmds.thrust * 4;

	signed long int kp_att = (signed long int)mikKopData.propGain*KP_ATT_MULT;
	signed long int kd_att = (signed long int)mikKopData.derivGain*KD_ATT_MULT;
	signed long int ki_att = (unsigned char)mikKopData.integGain*KI_ATT_MULT;
	signed long int kp_yaw = (signed long int)mikKopData.yawRateGain*KP_YAW_MULT;
	signed long int kp_yaw_acc = (signed long int)mikKopData.yawAccGain*KP_YAW_ACC_MULT;

	static signed long int prevTimeIntTerm=0;

	signed long int timeDiff = (signed long int) CountClockOverflow - prevTimeIntTerm;
	prevTimeIntTerm = CountClockOverflow;
	if(timeDiff < 0){
		timeDiff = 0;
	}

	static const signed long int scaleFactor=1000000;
	static const signed long int maxIntTerm=100000000;
	static const signed long int scaleFactorCommands=2500;  //10000 / 4
	//static const signed long int satValue=100;
	//static const signed long int yawSatValue=25;
	signed int satValue = mikKopData.satValue;
	signed int yawSatValue= mikKopData.yawsatValue;

	static signed long int  pitchIntTerm = 0, rollIntTerm = 0;


	//------------------- PITCH -------------------------//

	//pitch torque pid computation
	signed long int pitchTorque = kp_att*(desPitch*scaleFactorCommands - ((signed long int) mikKopData.pitch*1000))
	                              - kd_att*((signed long int) mikKopData.pitchRateLowPass*1000)
							      + pitchIntTerm; //TODO why no gain here?

	//pitch torque saturation and scaling
	CHECK_MIN_MAX(pitchTorque,-satValue*scaleFactor,satValue*scaleFactor);
	pitchTorque/=scaleFactor;

	//pitch integral term update
	pitchIntTerm+=(ki_att*(desPitch*scaleFactorCommands - ((signed long int) mikKopData.pitch*1000))*timeDiff)/(9700);

	//pitch integral term saturation
	CHECK_MIN_MAX(pitchIntTerm,-maxIntTerm,maxIntTerm);


	//------------------- ROLL -------------------------//

	//roll torque pid computation
	signed long int rollTorque =  kp_att*(desRoll*scaleFactorCommands - ((signed long int) mikKopData.roll*1000))
	                              - kd_att*((signed long int) mikKopData.rollRateLowPass*1000)
								  + rollIntTerm;

	//roll torque saturation and scaling
	CHECK_MIN_MAX(rollTorque,-satValue*scaleFactor,satValue*scaleFactor);
	rollTorque/=scaleFactor;

	//roll integral term update
	rollIntTerm+=(ki_att*(desRoll*scaleFactorCommands - ((signed long int) mikKopData.roll*1000))*timeDiff)/(9700);

	//roll i1ntegral term saturation
	CHECK_MIN_MAX(rollIntTerm,-maxIntTerm,maxIntTerm);


	//------------------- YAW -------------------------//
	//char print = 0;
	//yaw torque pid computation
	signed long int yawTorque=0;
	if(mikKopData.yawControlType==YAW_RATE){
		int desYawRate = (int) nearHovCmds.yawCmd;
		yawTorque = kp_yaw * ((signed long int)mikKopData.yawRateLowPass*1000 - (signed long int)desYawRate*10000)  ;    //sign is the opposite of the yaw angular velocity

/*		if ( mikKopData.timeMsMod30s%30 < 3){
			printf("yawTorque:%ld kpy:%ld desYawRate:%d yawRateLP:%d ", yawTorque, kp_yaw, desYawRate, mikKopData.yawRateLowPass);
			print = 1;
		}*/

	} else if(mikKopData.yawControlType==YAW_TORQUE){
		int desYawtorque = (int) nearHovCmds.yawCmd;
		yawTorque= kp_yaw_acc*desYawtorque*10000;
	}

	//yaw torque saturation and scaling
	CHECK_MIN_MAX(yawTorque,-yawSatValue*scaleFactor,yawSatValue*scaleFactor);
	yawTorque/=scaleFactor;

/*	if ( print){
		printf("yawTorque %ld rollTorque:%ld\r\n", yawTorque, rollTorque);
		print = 0;
	}*/
/*	printf("cmdsin: [%3u,%4d,%4d,%4d], state: [%6d,%6d,%6d,%6d], cmdout: [%7ld,%7ld,%7ld] \r\n",
			nearHovCmds.thrust,nearHovCmds.roll,nearHovCmds.pitch,nearHovCmds.yawCmd,
			mikKopData.roll, mikKopData.pitch, mikKopData.rollRateLowPass, mikKopData.pitchRateLowPass,
			desThrust, rollTorque, pitchTorque);*/
	//printf("ki_att=%5ld, pitchInt=%5ld, rollInt=%5ld\n\r", ki_att, pitchIntTerm, rollIntTerm);

		mikKopData.desThrust = (signed int)  desThrust;
		mikKopData.rollTorque = (signed int) rollTorque;
		mikKopData.pitchTorque = (signed int) pitchTorque;
		mikKopData.yawTorque = (signed int) yawTorque;

	// ----------------Motor commands ------------------//

	#ifdef IS_OCTOROTOR
	#warning "Building OctoRotor inputTransMatrix (Nr Motors = 8)!";
	static const unsigned int nrMotors = 8;
	static const signed char inputTransMatrix[8][4] = {
		{1,  8,  0,  8},
		{1,  8, -8, -8},
		{1,  0, -8,  8},
		{1, -8, -8, -8},
		{1, -8,  0,  8},
		{1, -8,  8, -8},
		{1,  0,  8,  8},
		{1,  8,  8, -8}
		};
	#else
	static const unsigned int nrMotors = 4;
	static const signed char inputTransMatrix[4][4] = {
		{1,  4,  0,  4},
		{1, -4,  0,  4},
		{1,  0, -4, -4},
		{1,  0,  4, -4}
		};
	#endif


        unsigned int i;
	for(i=0; i<nrMotors; i++)
	{
		signed long int tmp_sum;

		tmp_sum =  (desThrust * inputTransMatrix[i][0]);
		tmp_sum += (pitchTorque * inputTransMatrix[i][1]);
		tmp_sum += (rollTorque * inputTransMatrix[i][2]);
		tmp_sum += (yawTorque * inputTransMatrix[i][3]);

		tmp_sum /= nrMotors;


		//printf("tmp%c=%7ld, ", '1'+i, tmp_sum);
		CHECK_MIN_MAX(tmp_sum,MIN_MOT_CMD,MAX_MOT_CMD);

		//Motor[i].SetPoint = tmp_sum;

		motorBuf[i] = tmp_sum;
		//printf("tmps%c=%3ld, buf%c=%4u ", '1'+i, tmp_sum, '1'+i, motorBuf[i]);
	}

		mikKopData.motorBuf0 = (signed int) motorBuf[0];
		mikKopData.motorBuf1 = (signed int) motorBuf[1];
		mikKopData.motorBuf2 = (signed int) motorBuf[2];
		mikKopData.motorBuf3 = (signed int) motorBuf[3];
	//printf("\n\r");

}
